
#ifndef _CDROM_H_
  #define _CDROM_H_

  #ifdef __cplusplus
  extern "C" {
  #endif

  #define MEDIA_CH_CHANGED 0xFF
  #define MEDIA_CH_DONTKNOW 0
  #define MEDIA_CH_NOTCHANGED 1

  #define CD_INSTALLED 1
  #define CD_NOTINSTALLED -1

  #define AUDIO_TRACK 0
  #define DATA_TRACK 64

  extern int CD_installed;

  typedef struct {
    char min;
    char sec;
    char frame;
  } TCD_time;

  typedef struct {
    unsigned char control;
    unsigned char adr;
    unsigned char track;
    unsigned char index;

    TCD_time   trackTime;
    unsigned char zero;
    TCD_time   time;
  } TCD_playInfo ;

  typedef struct {
      unsigned char mode;
      unsigned char inputLeft;
      unsigned char volumeLeft;
      unsigned char inputRight;
      unsigned char volumeRight;
      unsigned char inputLeftMaster;
      unsigned char volumeLeftMaster;
      unsigned char inputRightMaster;
      unsigned char volumeRightMaster;
  } TCD_volumeInfo ;


  typedef struct {
    unsigned int doorOpened : 1;
    unsigned int doorUnlocked : 1;
    unsigned int supportsRawReading : 1;
    unsigned int writeable : 1;
    unsigned int canPlayAudioVideoTracks : 1;
    unsigned int supportsInterleaving : 1;
    unsigned int reserved : 1;
    unsigned int supportsPrefetchRequests : 1;
    unsigned int supportsAudioChannelControl : 1;
    unsigned int supportsRedBookAddressing : 1;
    unsigned int audioIsPlaying : 1;
    unsigned int reserved2 : 5;
  } TCD_status ;


  /*
  Error codes are the following:

    0  Write-protect violation
    1  Unknown unit
    2  Drive not ready
    3  Unknown command
    4  CRC error
    5  Bad drive request structure length
    6  Seek error
    7  Unknown media
    8  Sector not found
    9  Printer out of paper
    A  Write fault
    B  Read fault
    C  General failure
    D  Reserved
    E  Reserved
    F  Invalid disk change
  */

  typedef struct {
    unsigned int errorCode : 8;
    unsigned int done : 1;
    unsigned int busy : 1;
    unsigned int reserved : 5;
    unsigned int errorBit : 1;
  } TCD_error;




  typedef struct {
    unsigned short drives;
    unsigned char  firstDrive;

    unsigned char  tracksTotal;
    unsigned short trackCur;
    unsigned char  trackType;
    TCD_time       trackBegin;
    TCD_time       trackLength;

    TCD_time       audioStartTime;
    unsigned char  lowAudio;
    unsigned char  highAudio;
    TCD_time       diskLength;

    unsigned char  upc[7];
    unsigned char  diskid[6];
    TCD_status     status;
    unsigned short reservedStatus;
    TCD_error      error;
  } TCD_data ;




  //extern unsigned long CD_head_position (void);
  //extern void CD_seek (unsigned long location);
  //extern short CD_getupc (void); // zjistuje se automaticky

  extern void CD_getVolume (TCD_volumeInfo *vol);
  extern void CD_setVolume (TCD_volumeInfo *vol);

  extern void CD_setTrack (int tracknum);
  extern void CD_getStatus (void);

  extern void CD_play (TCD_time *begin, TCD_time *end);
  extern void CD_stop (void);
  extern void CD_resume (void);
  extern void CD_getPos (void);
  extern int  CD_mediaChange (void);
  extern void CD_open (void);
  extern void CD_close (void);
  extern void CD_eject (void);
  extern void CD_playTrack (int tracknum); // = -1
  extern int  CD_init (void);
  extern void CD_done (void);
  extern void CD_audioInit (void);
  extern void CD_lock (void);
  extern void CD_unlock (void);
  extern void CD_reset (void);

  extern TCD_data CD_info;
  extern TCD_playInfo CD_playInfo;

  #define CD_inserted ((!CD_info.error.errorCode) && (CD_info.diskid [0]) && (!CD_info.status.doorOpened))

  #ifdef __cplusplus
  }
  #endif



#endif
