
#include <dos.h>
#include <mem.h>
#include <process.h>
#include <stdlib.h>
#include <conio.h>

#include "dpmi.h"
#include "cgdk.h"
#include "svga.h"

#define DOS_INT 0x21
#define NET_INT 0x7A
#define SPOOL_INT 0x2F

#define DPMI_INT 0x31

#define DPMI_MAXAVAIL 0x00000500

#define DPMI_ALLOCDOSMEM 0x100
#define DPMI_FREEDOSMEM 0x101

#define DPMI_GETREALMODEVECTOR 0x0200
#define DPMI_SETREALMODEVECTOR 0x0201

#define DPMI_INTERRUPT 0x300

#define DPMI_NEWSELECTOR 0x0000
#define DPMI_SEGMENT2SELECTOR 0x0002

int DPMI_allocCount;
int DPMI_freeCount;
long DPMI_allocated;

unsigned long DPMI_maxAvail;

#ifndef __TURBOC__

  #include <i86.h>

  struct TDPMI_memInfo {
      unsigned LargestBlockAvail;
      unsigned MaxUnlockedPage;
      unsigned LargestLockablePage;
      unsigned LinAddrSpace;
      unsigned NumFreePagesAvail;
      unsigned NumPhysicalPagesFree;
      unsigned TotalPhysicalPages;
      unsigned FreeLinAddrSpace;
      unsigned SizeOfPageFile;
      unsigned Reserved[3];
  };

  static unsigned long DPMI_getMaxAvail (void)
  {
    union REGS regs;
    struct SREGS sregs = {0};
    TDPMI_memInfo memInfo;

    regs.x.eax = DPMI_MAXAVAIL;
    sregs.es = FP_SEG( &memInfo );
    regs.x.edi = FP_OFF( &memInfo );

    int386x( DPMI_INT, &regs, &regs, &sregs );
    return memInfo.LargestBlockAvail;
  }

  void DPMI_init (void)
  {
    if (DPMI_maxAvail == 0)
    {
      DPMI_maxAvail = DPMI_getMaxAvail ();
    }
  }

  typedef struct
  {
    unsigned short size;
    short selector;
    char reserved [12];
  } TDPMI_realBlock;

  void *DPMI_alloc (unsigned short Size)
  {
    union REGS regs;
    TDPMI_realBlock *memBlock;

    Size =_div (Size+sizeof (TDPMI_realBlock), 16);

    // DPMI call 100h allocates DOS memory
    regs.w.ax=DPMI_ALLOCDOSMEM;
    regs.w.bx= Size;
    int386 ( DPMI_INT, &regs, &regs);

    if (regs.x.cflag)
    {
      return NULL;
    }
    else
    {
      Size *= 16;
      DPMI_allocCount++;
      DPMI_allocated += Size;

      memBlock = (TDPMI_realBlock *) MK_DP (regs.w.ax, 0);
      memBlock->size = Size;
      memBlock->selector = regs.w.dx;

      return memBlock+1;
    }
  }

  void DPMI_free (void *DosPtr)
  {
    union REGS regs;
    TDPMI_realBlock *memBlock;

    memBlock = ((TDPMI_realBlock *) DosPtr) - 1;

    regs.w.ax = DPMI_FREEDOSMEM;
    regs.w.dx = memBlock->selector;

    int386 (DPMI_INT, &regs, &regs);
    if (regs.x.cflag)
    {
//      CGDK_exit ("DPMI free failed.\n", EXIT_FAILURE);
    }
    else
    {
      DPMI_freeCount++;
      DPMI_allocated-=memBlock->size;
    }
  }

  void DPMI_int (char IntNum, union RMINFO *rmi)
  {
      union REGS regs;
      static struct SREGS sregs;

     /* Use DMPI call 300h to issue the DOS interrupt */
      regs.w.ax = DPMI_INTERRUPT;
      regs.h.bl = IntNum;
      regs.h.bh = 0;
      regs.w.cx = 0;
      sregs.es = FP_SEG(rmi);
      regs.x.edi = FP_OFF(rmi);
      int386x ( DPMI_INT, &regs, &regs, &sregs );
  }

  short DPMI_newSelector (long address, long size)
  {
    union REGS regs;
    short sel;

    regs.w.ax = DPMI_NEWSELECTOR; // new selector
    regs.w.cx = 1; // only 1 selector
    int386 (DPMI_INT, &regs, &regs);
    sel = regs.w.ax;

    regs.w.ax = 0x800;
    regs.w.bx = hi (address);
    regs.w.cx = lo (address);
    regs.w.si = hi (size);
    regs.w.di = lo (size);
    int386 (DPMI_INT, &regs, &regs);

    regs.w.ax = 7;
    regs.w.dx = regs.w.cx;
    regs.w.cx = regs.w.bx;
    regs.w.bx = sel;
//    regs.w.cx = hi (address);
//    regs.w.dx = lo (address);
    int386 (DPMI_INT, &regs, &regs);

    regs.w.ax = 8;
    regs.w.bx = sel;
    regs.w.cx = hi (size);
    regs.w.dx = lo (size);
    int386 (DPMI_INT, &regs, &regs);

    regs.w.ax = 9;
    regs.w.bx = sel;
    regs.w.cx = 0x8092;
    int386 (DPMI_INT, &regs, &regs);

    return sel;
  }

  /*
  unsigned short DP_OFF (void *Ptr)
  {
    return (FP_OFF (Ptr) % 16);
  }

  unsigned short DP_SEG (void *Ptr)
  {
    return (FP_OFF (Ptr) / 16);
  }
  */

  void DPMI_debugInt_ (void);
  #pragma aux DPMI_debugInt_ = "int 3"

  void DPMI_debugInt (void)
  {
    DPMI_debugInt_ ();
  }

  long DPMI_getRealModeVector (int interruptNumber)
  {
    REGS regs={0};

    regs.x.eax = DPMI_GETREALMODEVECTOR;
    regs.x.ebx = interruptNumber;
    int386 (DPMI_INT, &regs, &regs);
    return (regs.x.ecx << 16) + regs.x.edx;
  }

  void DPMI_setRealModeVector (int interruptNumber, long addr)
  {
    REGS regs={0};

    regs.x.eax = DPMI_SETREALMODEVECTOR;
    regs.x.ebx = interruptNumber;
    regs.x.ecx = hi (addr);
    regs.x.edx = lo (addr);
    int386 (DPMI_INT, &regs, &regs);
  }


#else

  void DPMI_init (void)
  {
    DPMI_maxAvail = coreleft ();
  }

  void DPMI_int (char IntNum, RMINFO *rmi)
  {
    RMINFO _rmi;
    _rmi = *rmi;

    asm {
      push ax
      push bx
      push cx
      push dx
      push si
      push di
      push bp

      push ds
      push es
    }

    switch (IntNum)
    {
      case VGA_INT:
      {
        asm push bp;
        _AX = _rmi.x.ax;
        _BX = _rmi.x.bx;
        _CX = _rmi.x.cx;
        _DX = _rmi.x.dx;
        _SI = _rmi.x.si;
        _DI = _rmi.x.di;
        _ES = _rmi.x.es;
        _DS = _rmi.x.ds;
        _BP = _rmi.x.bp;
        asm {
          int VGA_INT;
          pop bp
        }
        break;
      }
      case DOS_INT:
      {
        asm push bp;
        _AX = _rmi.x.ax;
        _BX = _rmi.x.bx;
        _CX = _rmi.x.cx;
        _DX = _rmi.x.dx;
        _SI = _rmi.x.si;
        _DI = _rmi.x.di;
        _ES = _rmi.x.es;
        _DS = _rmi.x.ds;
        _BP = _rmi.x.bp;
        asm {
          int DOS_INT;
          pop bp
        }
        break;
      }
      case SPOOL_INT:
      {
        asm push bp;
        _AX = _rmi.x.ax;
        _BX = _rmi.x.bx;
        _CX = _rmi.x.cx;
        _DX = _rmi.x.dx;
        _SI = _rmi.x.si;
        _DI = _rmi.x.di;
	_ES = _rmi.x.es;
        _DS = _rmi.x.ds;
        _BP = _rmi.x.bp;
        asm {
          int SPOOL_INT;
          pop bp
        }
        break;
      }

      case NET_INT:
      {
        asm push bp;
        _AX = _rmi.x.ax;
        _BX = _rmi.x.bx;
        _CX = _rmi.x.cx;
        _DX = _rmi.x.dx;
        _SI = _rmi.x.si;
        _DI = _rmi.x.di;
        _ES = _rmi.x.es;
        _DS = _rmi.x.ds;
        _BP = _rmi.x.bp;
        asm {
          int NET_INT;
          pop bp
        }
        break;
      }
    }

    _rmi.x.ax = _AX;
    _rmi.x.ds = _DS;
    _rmi.x.es = _ES;
    _rmi.x.bx = _BX;
    _rmi.x.cx = _CX;
    _rmi.x.dx = _DX;
    _rmi.x.si = _SI;
    _rmi.x.di = _DI;

    asm {
      pop es
      pop ds

      pop bp
      pop di
      pop si
      pop dx
      pop cx
      pop bx
      pop ax
    }
    *rmi = _rmi;
  }


  void DPMI_disableInt (void)
  {
    asm cli;
  }

  void DPMI_enableInt (void)
  {
    asm sti;
  }

  /*
  unsigned short DOS_Alloc (unsigned short Size)
  {

    int a,b;
    unsigned segp;

    segp = -1;
    allocmem (Size / 16 + ((Size % 16) != 0), &segp);
    return segp;
  }
  */

  short DPMI_newSelector (long address, long size)
  {
    return 0;
  }

  void DPMI_debugInt (void)
  {
     asm int 3;
  }


#endif

#define REBOOT_VECT 0xF000FFF0
#define REBOOT_INT VGA_INT
#define REBOOT_AX 0x0003
#define REBOOT_TYPE 0x472


void DPMI_reboot (int warm)
{
  RMINFO rmi;

  // performs clean-up
  _cexit ();

  memset (&rmi, 0, sizeof rmi);
  rmi.x.eax = REBOOT_AX;
  *(long *) MK_DP (0,4*REBOOT_INT) = REBOOT_VECT;
  *(short *) MK_DP (0,REBOOT_TYPE) = (warm) ? REBOOT_WARM : REBOOT_COLD;
  DPMI_int (REBOOT_INT, &rmi);
  while (1)
    ;
}


static void DPMI_rebootWarm (void)
{
  DPMI_reboot (1);
}

static void DPMI_rebootCold (void)
{
  DPMI_reboot (0);
}


#ifdef __TURBOC__
int DPMI_getES (void)
{
  struct SREGS sregs;

  segread (&sregs);
  return sregs.es;
}

int DPMI_getDS (void)
{
  struct SREGS sregs;

  segread (&sregs);
  return sregs.ds;
}

void DPMI_setDS (int seg)
{
  _DS = seg;
}

void DPMI_setES (int seg)
{
  _ES = seg;
}

#else
int DPMI_getES_ (void);
#pragma aux DPMI_getES_ = \
" xor eax,eax"\
" mov ax,es"\
  value [eax]

int DPMI_getCS_ (void);
#pragma aux DPMI_getCS_ = \
" xor eax,eax"\
" mov ax,cs"\
  value [eax]

int DPMI_getDS_ (void);
#pragma aux DPMI_getDS_ = \
" xor eax,eax"\
" mov ax,ds"\
   value [eax]

void DPMI_setES_ (int seg);
#pragma aux DPMI_setES_ = \
"  mov  es,ax"\
   parm [eax]

void DPMI_setDS_ (int seg);
#pragma aux DPMI_setDS_ = \
"  mov  ds,ax"\
   parm [eax]

int DPMI_getES (void)
{
  return DPMI_getES_ ();
}

int DPMI_getDS (void)
{
  return DPMI_getDS_ ();
}

void DPMI_setDS (int seg)
{
  DPMI_setDS_ (seg);
}

void DPMI_setES (int seg)
{
  DPMI_setES_ (seg);
}

int DPMI_getCS (void)
{
  return DPMI_getCS_ ();
}

#endif
