
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <conio.h>
#include <process.h>
#include <ctype.h>

#include "cgdk.h"
#include "dpmi.h"
#include "ipx.h"
#include "ipxnode.h"
#include "timer.h"

#define NODE_DATAFRAGMAX 128      // maximum 512 * <this> data can be sended

#define NODE_SENDRETRY 2   // send data n+1-times (to make sure it will
//                               // be really delivered

#ifdef __TURBOC__
  #define NODE_PACKETMAX 32
#else
  #define NODE_PACKETMAX 256//(NODE_MAXNODES * NODE_DATAFRAGMAX * NODE_SENDDATARETRY)
                                // 0 packet - local
#endif

#define NODE_FREE 0             // means that the node socket is free
#define NODE_LOGGED 1           // means that the node socket is used
#define PACKET_CERBERUS 3080760320// packet stamp used for checking incoming packets

typedef struct
{
        TECBHeader     ECB;           // event control block used for sending or receiving
        TIPXHeader     IPX;           // ipx header
        long           Type;          // type stamp (should be PACKET_CERBERUS)
        long           FragTotal;     // total fragments of s/r data (n)
        long           Frag;          // number of this sended fragment (0 .. n-1)
        long           DataSizeTotal; // total number of s/r bytes
        long           DataSize;      // size of data in this packet
        long           TimeStamp;     // local time of sended packet
        long           DataType;      // type of sended/received data
        char           reserved [6];
        char           Data [NODE_DATAFRAGSIZE];  // user data
} TPacket;

typedef struct
{
  char       Status;           // free or logged
  long       TimeStampOut [NODE_DATAMAX];     // the last time stamp
  long       TimeStampIn [NODE_DATAMAX];
  long       nextTimeStamp [NODE_DATAMAX];
  long       waiting [NODE_DATAMAX];
  TNodeAddr  router;
  TIPXAddress Address;          // IPX address of node
} TNode;

#define NODE_EXTRASERVER_MAX 10

TIPXAddress NODE_extraServer [NODE_EXTRASERVER_MAX];
int NODE_extraServerCount;


int NODE_packetLostDelay=NODE_PACKETLOSTDELAY_DEFAULT;


unsigned short IPX_socket;              // ipx socket used for sending and receiving

static long Frag [NODE_DATAFRAGMAX];       // index of frags in the Packet []

static TNode Node [NODE_MAXNODES];         // info about all nodes

TIPXAddress LocalAddress;            // IPX address of the local machine
static TIPXAddress BroadcastAddress;

static TPacket *LocalPacket;               // the local sending packet
static TPacket *Packet [NODE_PACKETMAX];   // all network receiving packets


static long waiting;

static int NODE_socketOpened;


static char *hex = "0123456789ABCDEF";

static void PrintsAddress (unsigned char Address [], char *Str, int Len)
{
  int i;

  for (i=0 ; i<Len ; i++)
  {
    *Str++ = hex[Address [i]>>4];
    *Str++ = hex[Address [i]&0x0F];
  }
  *Str = 0;
}

static char Scans (char a)
{
  a = tolower (a);

  if (a >= 'a')
  {
    a -= 'a'-10;
  }
  else
  {
    a -= '0';
  }

  return a;
}

static void ScansAddress (unsigned char Address [], char *Str)
{
  int i;
  int c;
  char a,b;

  i = 0;
  while (a = *Str++, a != 0)
  {
    a = Scans (a);
    b = Scans (*Str++);
    Address [i] = (a << 4) + b;
    i++;
  }
}

int NODE_addExtraServer (TNetAddr *addr)
{
  short time;

  if (NODE_extraServerCount < NODE_EXTRASERVER_MAX)
  {
    memcpy (&NODE_extraServer [NODE_extraServerCount].Network, addr, sizeof (TNetAddr));
    memset (&NODE_extraServer [NODE_extraServerCount].Node, 0, sizeof (TNodeAddr));
    if (IPX_GetLocalTarget (&NODE_extraServer [NODE_extraServerCount], &NODE_extraServer [NODE_extraServerCount].Node, &time))
    {
      NODE_extraServerCount++;
      return 0;
    }
  }
  return 1;
}

void NODE_printsNodeAddress (TIPXAddress *Address, char *Str)
{
  PrintsAddress (Address->Node, Str, 6);
}

void NODE_printsNetAddress (TIPXAddress *Address, char *Str)
{
  PrintsAddress (Address->Network, Str, 4);
}

void NODE_scansNodeAddress (TIPXAddress *Address, char *Str)
{
  ScansAddress (Address->Node, Str);
}

void NODE_scansNetAddress (TIPXAddress *Address, char *Str)
{
  ScansAddress (Address->Network, Str);
}

void NODE_printsSocketAddress (TIPXAddress *Address, char *Str)
{
  char c;

  PrintsAddress ((unsigned char *)&Address->Socket, Str, 2);
//  c = Str [1];
//  Str [1] = Str [0];
//  Str [0] = c;
}

void NODE_scansSocketAddress (TIPXAddress *Address, char *Str)
{
  char c;
  unsigned char *s = (unsigned char *) &Address->Socket;

  ScansAddress (s, Str);
//  c = s [1];
//  s [1] = s [0];
//  s [0] = c;
}


void NODE_SetExportTime (int NodeNumber, long NewTime)
{
  int i;
  for (i = 0; i < NODE_DATAMAX; i++)
    Node [NodeNumber].TimeStampOut [i] = NewTime;
}

void NODE_SetImportTime (int NodeNumber, long NewTime)
{
  int i;
  for (i = 0; i < NODE_DATAMAX; i++)
    Node [NodeNumber].TimeStampIn [i] = NewTime;
}

void NODE_setWaiting (int NodeNumber, long NewTime)
{
  int i;
  for (i = 0; i < NODE_DATAMAX; i++)
    Node [NodeNumber].waiting [i] = NewTime;
}


void NODE_setNextTimeStamp (int NodeNumber, long NewTime)
{
  int i;
  for (i = 0; i < NODE_DATAMAX; i++)
    Node [NodeNumber].nextTimeStamp [i] = NewTime;
}

int NODE_New (TIPXAddress *Address)
{
  int i = 0;
  short time;

  // check out if there is a free node
  while ((Node [i] . Status != NODE_FREE) && (i < NODE_MAXNODES))
    i++;
  if (i > NODE_MAXNODES)
    // all nodes are full
    return -1;
  // there is a free node
  Node [i] . Status = NODE_LOGGED;
  Node [i] . Address = *Address;
  Node [i] . Address.Socket = IPX_socket;
  IPX_GetLocalTarget (Address, &Node [i] . router, &time);
  NODE_SetImportTime (i, 0);
  NODE_SetExportTime (i, 0);
  NODE_setWaiting (i, 0);
  NODE_setNextTimeStamp (i, 0);

  // return the node number
  return i;
}


void NODE_Delete (int NodeNumber)
{
  // check if the node is logged
  if (Node [NodeNumber] . Status != NODE_LOGGED)
    return;
  Node [NodeNumber] . Status = NODE_FREE;
}





void Packet_Reset (int packet)
{
  if ((packet != 0) && (Packet  [packet]->ECB.InUseFlag == 0))
  {
    Packet [packet]->Type = -1;
    IPX_ReceivePacket (&Packet [packet]->ECB);
  }
}

void Packet_Cancel (TPacket *packet)
{
//  while (packet->ECB.InUseFlag)
//    ;
//
  packet->Type = -1;

  while (IPX_CancelEvent (&packet->ECB) == 0xF9)
    ;
}


void NODE_ClearWastePackets (void)
{
  int i, j;

  for (i = 1; i < NODE_PACKETMAX; i++)
  {
    if ((Packet [i]->ECB.InUseFlag == 0) && (Packet [i]->DataType))
    {
      for (j = 0; j < NODE_MAXNODES; j++)
      {
        if ((Node [j].Status == NODE_LOGGED) &&
            (memcmp ((void *) &Node [j].Address, (void *) &Packet [i]->IPX.Src, sizeof (TIPXAddress)) == 0))
        {
          break;
        }
      }
      if (j == NODE_MAXNODES)
      {
        Packet_Reset (i); /// i or j ?
      }
    }
  }
}


TIPXAddress *NODE_GetLocalAddress (void)
{
  return &LocalAddress;
}


TIPXAddress *NODE_GetNodeAddress (int NodeNumber)
{
  if (NodeNumber <= NODE_MAXNODES)
    return &Node [NodeNumber].Address;
  else
    return NULL;
}

TPacket *NODE_NewPacket (void)
{
  TPacket *Packet;

  // allocate ECB in low DOS memory
  Packet = (TPacket *) DPMI_alloc (sizeof (TPacket));
  if (Packet != NULL)
  {
    memset (Packet, 0, sizeof (TPacket));

    // fill the IPX header info
    Packet->IPX.Src.Socket = Packet->IPX.Dst.Socket = IPX_socket;

    // fill the ECB fragments info
    Packet->ECB.SocketNumber = IPX_socket;
    Packet->ECB.FragCount = 1;
    Packet->ECB.FragOff = DP_OFF (Packet) + sizeof (TECBHeader);
    Packet->ECB.FragSeg = DP_SEG (Packet);
    Packet->ECB.FragSize = sizeof (TPacket) - sizeof (TECBHeader);
    return Packet;
  }
  return NULL;
}



int NODE_Init (void)
{
  int i;

  if (sizeof (TPacket) - sizeof (TECBHeader) == 576)
    ;
  else
    return -1;


  if (IPX_socket == 0)
  {
    IPX_socket = IPX_DEFAULTSOCKET;
  }

  // check first if the IPX is installed
  if (IPX_Initialize ())
  {
    // allocate a socket for sending and receiving
    if (IPX_OpenSocket (IPX_socket))
    {
      NODE_socketOpened = 1;
//      IPXSocket = Socket;

      memset (BroadcastAddress.Network, 0, 4);
      memset (BroadcastAddress.Node, 0xFF, 6);

      // clear all nodes
      memset (Node, 0, sizeof Node);

      // get local console address
      IPX_GetLocalAddress (&LocalAddress);
      LocalAddress.Socket = IPX_socket;

      // create the broadcast node
      NODE_New (&BroadcastAddress);

      // init the local packet
      LocalPacket = NODE_NewPacket ();
      // set it to recognize that it's the right paket
      LocalPacket->Type = PACKET_CERBERUS;
      // init packets and ecbs
      for (i=0 ; i<NODE_PACKETMAX ; i++)
      {
        // prepare for receiving except packet 0
        if (((Packet [i] = NODE_NewPacket ()) != NULL))
        {
          Packet_Reset (i);
        }
        else
        {
          NODE_Done ();
          return 0;
        }
      }

      return 1;
    }
  }
  return 0;
}

void NODE_FlushPackets (void)
{
  int i;

  for (i = 0; i < NODE_PACKETMAX; i++)
    Packet_Reset (i);
}


void NODE_Done (void)
{
  int i;

  if (LocalPacket)
  {
    Packet_Cancel (LocalPacket);
    DPMI_free (LocalPacket);
    LocalPacket = NULL;
  }

  // free ecb and packet buffers from dos memory
  for (i=0 ; i<NODE_PACKETMAX ; i++)
  {
    // free ECB structures
    if (Packet [i])
    {
      Packet_Cancel (Packet [i]);
      DPMI_free (Packet [i]);
      Packet [i] = NULL;
    }
  }

  // deinitialize IPX
  if (NODE_socketOpened)
  {
    IPX_CloseSocket (IPX_socket);
    NODE_socketOpened = 0;
  }

//  IPX_DisableCPUControl ();

  IPX_done ();
}


// A destination of NodeNumber == NODE_BROADCAST means that data will be
// sended to all nodes.

int NODE_Send (int NodeNumber, void *Data, int DataSize, int DataType)
{
  int i,j;
  // count fragments total
  int FragTotal = (DataSize / NODE_DATAFRAGSIZE) + (((DataSize % NODE_DATAFRAGSIZE) != 0));
  int DataRetry;
  short time;

  if (NodeNumber == NODE_BROADCAST)
  {
    abs (1);
  }

  if (DataType == 0)
  {
    if (DataSize > NODE_DATAFRAGSIZE) // broadcast
      return 0;
    else
      DataRetry = 0;
  }
  else
  {
    if (DataType == 1)
    {
      DataRetry = NODE_SENDRETRY;
    }
    else
    {
      DataRetry = 0;
    }
  }

  // send data n-times to make sure data will be sended
  for (j = 0; j <= DataRetry; j++)
  {
    if (// use only opened (active) nodes
       (Node [NodeNumber].Status == NODE_LOGGED) &&
    // do not receive more fragments than DATAFRAGMAX
       (FragTotal <= NODE_DATAFRAGMAX))
    {
      // store fragments total
      LocalPacket->FragTotal = FragTotal;
      // store the timestamp
      LocalPacket->TimeStamp = Node [NodeNumber].TimeStampOut [DataType];
      // store the sended data type
      LocalPacket->DataType = DataType;
      // copy destination address to ipx structure
      memcpy (&LocalPacket->IPX.Dst, &Node [NodeNumber].Address, sizeof (TIPXAddress));

//      IPX_GetLocalTarget (&LocalPacket->IPX.Dst, &LocalPacket->ECB.ImmediateAddress, &time);
//      memcpy (LocalPacket->ECB.ImmediateAddress, Node [NodeNumber].Address.Node, sizeof (TNodeAddr));
      memcpy (LocalPacket->ECB.ImmediateAddress, Node [NodeNumber].router, sizeof (TNodeAddr));

      // send all fragments
      for (i = 0; i < LocalPacket->FragTotal; i++)
      {
        // fill the current fragment count
        LocalPacket->Frag = i;
        // fill the packet time stamp

        // copy data to send to packets
        if (i == ((LocalPacket->FragTotal)-1))
        {
          // copy the rest
          memcpy (LocalPacket->Data, (char *) (Data)+i*NODE_DATAFRAGSIZE,
                  DataSize - i * NODE_DATAFRAGSIZE);
        }
        else
        {
          // copy all frag data
          memcpy (LocalPacket->Data, (char *) (Data)+i*NODE_DATAFRAGSIZE,
                  NODE_DATAFRAGSIZE);
        }


        // wait till ipx is using ECB (shouldn't take too long)
//         while (LocalPacket->ECB.InUseFlag) ;
        IPX_SendPacket (&LocalPacket->ECB);// && (!LocalPacket->ECB.CompletitionFlag);

        if (NodeNumber == NODE_BROADCAST)
        {
          for (i = 0; i < NODE_extraServerCount; i++)
          {
            memcpy (LocalPacket->IPX.Dst.Network, NODE_extraServer [i].Network, sizeof (TNetAddr));
            // copy destination node to immediate address

            memcpy (LocalPacket->ECB.ImmediateAddress, NODE_extraServer [i].Node, sizeof (TNodeAddr));


            IPX_SendPacket (&LocalPacket->ECB);// && (!LocalPacket->ECB.CompletitionFlag);

//            while (LocalPacket->ECB.InUseFlag) ;
          }
        }
//        else
          // call the ipx function to send the packet via ecb


      }

      // increase the local machine time stamp
      if (DataType)
      {
        Node [NodeNumber].TimeStampOut [DataType]++;
      }
    }
    return 1; // some or all data packets has been sended
  }

  // return an error
  return 0;
}


int NODE_Receive (int NodeNumber, void *Data, int DataSize, int DataType)
{
  int i,j;
  long FragCur = 0;
  long FragTotal = (DataSize / NODE_DATAFRAGSIZE) + (((DataSize % NODE_DATAFRAGSIZE) != 0));
  long TimeStamp;
  long time;

  if (// use only opened (active) nodes
     (Node [NodeNumber].Status == NODE_LOGGED) &&
     // do not receive more fragments than DATAFRAGMAX
     (FragTotal <= NODE_DATAFRAGMAX))
  {
    TimeStamp = Node [NodeNumber].TimeStampIn [DataType];
    // check all -receiving- packets and ecbs

    a:

    for (i = 1 ; (i < NODE_PACKETMAX); i++)
    {
      // check if IPX is using current ECB
      if (Packet [i]->ECB.InUseFlag == 0)
      {
        // check if the packet had commed from the address specified by NodeNumber
        // or it is a broadcast packet
        if ( (Packet [i]->Type == PACKET_CERBERUS) &&
          // check if the incoming packet is OK
          (Packet [i]->ECB.CompletitionFlag == 0) &&
          // do not receive data from self
          (memcmp ((void *) &Packet [i]->IPX.Src, (void *) &LocalAddress, sizeof (TIPXAddress)) != 0))
        {
          // check the address
          if (((memcmp ((void *) &Packet [i]->IPX.Src, (void *) &Node [NodeNumber].Address, sizeof (TIPXAddress)) == 0) ||
          (NodeNumber == NODE_BROADCAST)) && (Packet [i]->DataType == DataType))
          {
            // this packet is coming from the desired node
            // check if the incoming packet is newer than the time stamp and the best packet

            if (Packet [i]->TimeStamp >= TimeStamp)
            // we received packet
            {
              if (Packet [i]->TimeStamp == TimeStamp)
              {
                for (j = 0;
                    (j < FragTotal) &&
                    ( (Packet [Frag [j]]->Frag != Packet [i]->Frag) ||
                      (Frag [j] == 0));
                    j++)
                  ;

                if (j == FragTotal)
                {
                  // store the number of the packet in the fragment list
                  Frag [Packet [i]->Frag] = i;
                  // proccess the next fragment
                  FragCur++;
//                if (DataType)
//                {
//                  break;
//                }
                }
                else
                {
                  Packet_Reset (i);
                }
              }
              else
              {
                if (Node [NodeNumber].nextTimeStamp [DataType] > Packet [i]->TimeStamp)
                {
                  Node [NodeNumber].nextTimeStamp [DataType] = Packet [i]->TimeStamp;
                }
                time = TIMER_timeTotal;
                if (Node [NodeNumber].waiting [DataType])
                {
                  if ((time - Node [NodeNumber].waiting [DataType]) >= NODE_packetLostDelay)
                  {
                    Node [NodeNumber].TimeStampIn [DataType] = Node [NodeNumber].nextTimeStamp [DataType];
                    Node [NodeNumber].waiting [DataType] = 0;
                    Node [NodeNumber].nextTimeStamp [DataType]= MAXLONG;
                    goto a;
                  }
                }
                else
                {
                  Node [NodeNumber].waiting [DataType] = time;
                }
              }
            }
            else
              Packet_Reset (i);
          }
        }
        else
          Packet_Reset (i);
        // this packet is unknown, bad or outdated so we better receieve a new one
      }
    }

    // check if all fragments has been collected
    if (FragCur == FragTotal)
    {
      // check the data type
      if (DataType)
      {
        Node [NodeNumber].TimeStampIn [DataType]++;
        Node [NodeNumber].waiting [DataType] = 0;
        Node [NodeNumber].nextTimeStamp [DataType]= MAXLONG;
      }
      // copy out the data

      for (i = 0; i < FragTotal; i++)
      {
        if (i == (FragTotal-1))
        {
          j = DataSize - i*NODE_DATAFRAGSIZE;
        }
        else
        {
          j = NODE_DATAFRAGSIZE;
        }
        memcpy ((char *)Data+i*NODE_DATAFRAGSIZE, Packet [Frag [i]]->Data, j);

        Packet_Reset (Frag [i]);
        Frag [i] = 0;
      }
      // data successfully received
      return 1;
    }
  }
  return 0;
}

unsigned short NODE_GetCounter (void)
{
  return IPX_GetIntervalMarker ();
}

void NODE_Delay (int ms)
{
  unsigned short t = NODE_GetCounter ();
  unsigned short t2;

  do
  {
    while (t2 = NODE_GetCounter (), t2 == t)
      ;
    t = t2;
    ms--;
  }
  while (ms);
}
