
#include <stdlib.h>
#include <math.h>

#include "cgdk.h"
#include "svga.h"
#include "l256_.h"
#include "lin.h"
#include "timer.h"
#include "dpmi.h"
#include "keyb.h"

#define L256_MOVWIN2 0
#define L256_MOVSPR2 1
#define LIN_MOVWIN 2
#define LIN_MOVSPR 3

struct TL256C
{
  void *dst;
  long dstFormatX;
  void *spr;
  short  sprFormatX;
  short  sizeX;
  short  sizeY;
  short  type;
};

static TL256C *L256C;
static int L256C_last;
static int L256C_maxCached;

void L256C_done (void)
{
  if (L256C)
  {
    _free (L256C);
    L256C = NULL;
  }
  L256C_maxCached = 0;
  L256C_last = 0;
}

void L256C_init (void)
{
  L256C_done ();
  L256C_maxCached = SVGAC_maxCached;
  L256C = (TL256C *) _alloc (L256C_maxCached * sizeof (TL256C));
  L256C_last = 0;
}

void L256C_putSpr (void *dst, int dstFormatX,
                          void *spr, int sprFormatX, int sizeX, int sizeY)
{
  int pos;

  if (sizeY)
  {
    pos = L256C_last;
    L256C_last++;
    L256C [pos].dst = dst;
    L256C [pos].dstFormatX = dstFormatX;
    L256C [pos].spr = spr;
    L256C [pos].sprFormatX = sprFormatX;
    L256C [pos].sizeX = sizeX;
    L256C [pos].sizeY = sizeY;
    L256C [pos].type = LIN_MOVSPR;
  }
}

void L256C_putWin (void *dst, int dstFormatX,
                          void *spr, int sprFormatX, int sizeX, int sizeY)
{
  int pos;

  if (sizeY)
  {
    pos = L256C_last;
    L256C_last++;
    L256C [pos].dst = dst;
    L256C [pos].dstFormatX = dstFormatX;
    L256C [pos].spr = spr;
    L256C [pos].sprFormatX = sprFormatX;
    L256C [pos].sizeX = sizeX;
    L256C [pos].sizeY = sizeY;
    L256C [pos].type = LIN_MOVWIN;
  }
}

void L256C_putWin2 (void *dst, void *dst2,
                          void *spr, int sprFormatX, int sizeX, int sizeY)
{
  int pos;

  if (sizeY)
  {
    pos = L256C_last;
    L256C [pos].dst = dst;
    L256C [pos].dstFormatX = *(long *) &dst2;
    L256C [pos].spr = spr;
    L256C [pos].sprFormatX = sprFormatX;
    L256C [pos].sizeX = sizeX;
    L256C [pos].sizeY = sizeY;
    L256C [pos].type = L256_MOVWIN2;
    L256C_last++;
  }
}


void L256C_putSpr2 (void *dst, void *dst2,
                          void *spr, int sprFormatX, int sizeX, int sizeY)
{
  int pos;

  if (sizeY)
  {
    pos = L256C_last;
    L256C_last++;
    L256C [pos].dst = dst;
    L256C [pos].dstFormatX = *(long *) &dst2;;
    L256C [pos].spr = spr;
    L256C [pos].sprFormatX = sprFormatX;
    L256C [pos].sizeX = sizeX;
    L256C [pos].sizeY = sizeY;
    L256C [pos].type = L256_MOVSPR2;
  }
}



static void L256C_flushMov (TL256C *lc)
{
  switch (lc->type)
  {
    #ifndef __TURBOC__
    case L256_MOVWIN2:
    {
      L256_movWin2_ (lc->dst, (void *) lc->dstFormatX, lc->spr, lc->sprFormatX, lc->sizeX, lc->sizeY);
      break;
    }
    case L256_MOVSPR2:
    {
      L256_movSpr2_ (lc->dst, (void *) lc->dstFormatX, lc->spr, lc->sprFormatX, lc->sizeX, lc->sizeY);
      break;
    }
    #endif
    case LIN_MOVWIN:
    {
      LIN_movWin_ (lc->dst, lc->dstFormatX, lc->spr, lc->sprFormatX, lc->sizeX, lc->sizeY);
      break;
    }
    case LIN_MOVSPR:
    {
      LIN_movSpr_ (lc->dst, lc->dstFormatX, lc->spr, lc->sprFormatX, lc->sizeX, lc->sizeY);
      break;
    }
  }
}

void L256C_flush2 (void)
{
  int i;
  int retraceTime;
  int retraceDetect;
  long beamY;
  long beamMaxY;
  long end1Y, end2Y;
//  long end1X, end2X;
  TL256C *lc;
  int es;
  double retraceRest;
  double beamRest;

  lc = L256C;

  if (SVGA_linearSeg)
  {
    es = DPMI_getES ();
    DPMI_setES (SVGA_linearSeg);
  }
  else
  {
    es = 0;
  }

  if (SVGA_enableRetrace)
  {
    i = 0;

    retraceTime = TIMER_retraceTime;
    retraceDetect = 0;
    while (retraceRest = fmod (TIMER_retraceTime, TIMER_retraceTimeMaxF),
	   retraceRest < 0.95*TIMER_retraceTimeMaxF)
    {
      beamRest = (retraceRest-0.1*TIMER_retraceTimeMaxF) / TIMER_retraceTimeMaxF;
      beamY = SVGA_curYPrev + SVGA_maxY * beamRest;
      beamMaxY = SVGA_curYPrev + SVGA_maxY;

      retraceTime = TIMER_retraceTime;
      while ((TIMER_retraceTime == retraceTime) && (i < L256C_last) &&
	     (retraceDetect = SVGA_retraceDetect (), !retraceDetect) )
      {
	end1Y = ((long)lc->dst) / SVGA_virtualX + lc->sizeY;
	end2Y = ((long)lc->dstFormatX) / SVGA_virtualX + lc->sizeY;

	if ( ((end1Y < beamY) || (end1Y > beamMaxY)) &&
	     ( (lc->type == LIN_MOVWIN) || (lc->type == LIN_MOVSPR)||(end2Y < beamY) || (end2Y > beamMaxY)
	     )
	   )
	{
	  L256C_flushMov (lc);
	  *(int *) &lc += sizeof (TL256C);
	  i++;
	}
      }
      if (i == L256C_last || retraceDetect)
      {
	break;
      }
    }

    if (i == L256C_last)
    {
      SVGA_wait (1);
    }
    else
    {
      retraceTime = TIMER_retraceTime;
      while (TIMER_retraceTime - retraceTime < (0.1*TIMER_retraceTimeMaxF))
      {
	if (SVGA_retraceDetect ())
	{
	  SVGA_wait (1);
	  break;
	}
      }
    }
  }
  else
    SVGA_wait (1);

  for ( ; i < L256C_last; i++)
  {
    L256C_flushMov (lc);
    *(int *) &lc += sizeof (TL256C);
  }

  SVGA_scroll ();

  L256C_last = 0;

  if (es)
    DPMI_setES (es);
}


void L256C_flush (void)
{
  int i;
  long beamY;
  long beamMaxY;
  long end1Y, end2Y;
  TL256C *lc;
  int es;

  lc = L256C;

  if (SVGA_linearSeg)
  {
    es = DPMI_getES ();
    DPMI_setES (SVGA_linearSeg);
  }
  else
  {
    es = 0;
  }

  i = 0;

  if (SVGA_enableRetrace)
  {
    beamY = SVGA_curYPrev + SVGA_maxY/3;
    beamMaxY = SVGA_curYPrev + SVGA_maxY;

    while (fmod (TIMER_retraceTime, TIMER_retraceTimeMaxF) < 0.666*TIMER_retraceTimeMaxF)
      ;

    while (i < L256C_last)
    {
      end1Y = ((long)lc->dst) / SVGA_virtualX + lc->sizeY;
      end2Y = ((long)lc->dstFormatX) / SVGA_virtualX + lc->sizeY;

      if ( ((end1Y < beamY) || (end1Y > beamMaxY)) &&
	     ( (lc->type == LIN_MOVWIN) || (lc->type == LIN_MOVSPR)||(end2Y < beamY) || (end2Y > beamMaxY)
	     )
	   )
      {
	L256C_flushMov (lc);
	*(int *) &lc += sizeof (TL256C);
	i++;
      }
      else
      {
	break;
      }
    }
  }


  SVGA_wait (1);

  for ( ; i < L256C_last; i++)
  {
    L256C_flushMov (lc);
    *(int *) &lc += sizeof (TL256C);
  }

  L256C_last = 0;

  if (es)
    DPMI_setES (es);
}
