
#define AC_INDEX        0x3c0   // Attribute controller index register
#define MISC_OUTPUT     0x3c2   // Miscellaneous Output register
#define SC_INDEX        0x3c4   // Sequence Controller Index
#define GC_INDEX        0x3ce   // Graphics controller Index
//#define CRTC_INDEX      0x3d4   // CRT Controller Index

////////////////////////////////////////////////////////////////////////
// The following sets of arrays are the register parameters to set up //
// the various x modes.                                               //
////////////////////////////////////////////////////////////////////////

#define MAX_CRTC_REGISTER 20

struct TXSetup {
  int width;
  int height;
  int dotClock;
  int registers [MAX_CRTC_REGISTER];
};

static TXSetup X256Y200 = {
  256,
  200,
  0x0e3,
  {
    0x5f00,   //horz total
    0x3f01,   //horz displayed
    0x4202,   //start horz blanking
    0x9f03,   //end horz blanking
    0x4c04,   //start h sync
    0x0005,   //end h sync
    0x0014,   //turn off dword mode
    0xe317   //turn on byte mode
  }
};

static TXSetup X256Y240 = {
  256,
  240,
  0x0e3,
  {
    0x5f00,  //horz total
    0x3f01,  //horz displayed
    0x4202,  //start horz blanking
    0x9f03,  //end horz blanking
    0x4c04,  //start h sync
    0x0005,  //end h sync
    0x0d06,  //vertical total
    0x3e07,  //overflow (bit 8 of vertical counts)
    0x4109,  //cell height (2 to double-scan)
    0xea10,  //v sync start
    0xac11,  //v sync end and protect cr0-cr7
    0xdf12,  //vertical displayed
    0x0014,  //turn off dword mode
    0xe715,  //v blank start
    0x0616,  //v blank end
    0xe317   //turn on byte mode
  }
};


static TXSetup X256Y224 = {
  256,
  224,
  0x0e3,
  {
    0x12e3,
    0x5f00,
    0x3f01,
    0x4202,
    0x8203,
    0x4a04,
    0x9a05,
    0x0b06,
    0x3e07,
    0x0008,
    0x4109,
    0xda10,
    0x9c11,
    0xbf12,
    0x2013,
    0x0014,
    0xc715,
    0x0416,
    0xe317
  }
};

static TXSetup X256Y256 = {
  256,
  256,
  0x0e3,
  {
    0x5f00,
    0x3f01,
    0x4002,
    0x8203,
    0x4a04,
    0x9a05,
    0x2306,
    0xb207,
    0x0008,
    0x6109,
    0x0a10,
    0xac11,
    0xff12,
    0x2013,
    0x0014,
    0x0715,
    0x1a16,
    0xe317
  }
};

static TXSetup X256Y400 = {
  256,
  400,
  0x0e3,
  {
    0x5f00,  //horz total
    0x3f01,  //horz displayed
    0x4202,  //start horz blanking
    0x9f03,  //end horz blanking
    0x4c04,  //start h sync
    0x0005,  //end h sync
    0x4009,  //cell height
    0x0014,  //turn off dword mode
    0xe317   //turn on byte mode
  }
};

static TXSetup X256Y480 = {
  256,
  480,
  0x0e3,
  {
    0x5f00,  //horz total
    0x3f01,  //horz displayed
    0x4202,  //start horz blanking
    0x9f03,  //end horz blanking
    0x4c04,  //start h sync
    0x0005,  //end h sync
    0x0d06,  //vertical total
    0x3e07,  //overflow (bit 8 of vertical counts)
    0x4009,  //cell height (2 to double-scan)
    0xea10,  //v sync start
    0xac11,  //v sync end and protect cr0-cr7
    0xdf12,  //vertical displayed
    0x0014,  //turn off dword mode
    0xe715,  //v blank start
    0x0616,  //v blank end
    0xe317   //turn on byte mode
  }
};


static TXSetup X320Y200 = {
  320,
  200,
  0,
  {
    0x0014, //Turn off dword mode
    0xe317  //Turn on byte mode
  }
};

static TXSetup X320Y240 = {
  320,
  240,
  0x0e3,
  {
    0x0d06,  //vertical total
    0x3e07,  //overflow (bit 8 of vertical counts)
    0x4109,  //cell height (2 to double-scan)
    0xea10,  //v sync start
    0xac11,  //v sync end and protect cr0-cr7
    0xdf12,  //vertical displayed
    0x0014,  //turn off dword mode
    0xe715,  //v blank start
    0x0616,  //v blank end
    0xe317   //turn on byte mode
  }
};

static TXSetup X360Y200 = {
  360,
  200,
  0x0e7,
  {
    0x6b00,  //horz total
    0x5901,  //horz displayed
    0x5a02,  //start horz blanking
    0x8e03,  //end horz blanking
    0x5e04,  //start h sync
    0x8a05,  //end h sync
    0x0014,  //turn off dword mode
    0xe317   //turn on byte mode
  }
};

static TXSetup X360Y240 = {
  360,
  240,
  0x0e7,
  {
    0x6b00,  //horz total
    0x5901,  //horz displayed
    0x5a02,  //start horz blanking
    0x8e03,  //end horz blanking
    0x5e04,  //start h sync
    0x8a05,  //end h sync
    0x0d06,  //vertical total
    0x3e07,  //overflow (bit 8 of vertical counts)
    0x4109,  //cell height (2 to double-scan)
    0xea10,  //v sync start
    0xac11,  //v sync end and protect cr0-cr7
    0xdf12,  //vertical displayed
    0x2d13,  //offset;
    0x0014,  //turn off dword mode
    0xe715,  //v blank start
    0x0616,  //v blank end
    0xe317   //turn on byte mode
  }
};


static TXSetup X376Y282 = {
  376,
  282,
  0x0e7,
  {
    0x6e00,  //horz total
    0x5d01,  //horz displayed
    0x5e02,  //start horz blanking
    0x9103,  //end horz blanking
    0x6204,  //start h sync
    0x8f05,  //end h sync
    0x6206,  //vertical total
    0xf007,  //overflow
    0x6109,  //cell height
    0x310f,  //Unknown?-->VPutz
    0x3710,  //v sync start
    0x8911,  //v sync end and protect cr0-cr7
    0x3312,  //vertical displayed
    0x2f13,  //offset
    0x0014,  //turn off dword mode
    0x3c15,  //v blank start
    0x5c16,  //v blank end
    0xe317  //turn on byte mode
  }
};


static TXSetup X320Y400 = {
  320,
  400,
  0x0e3,
  {
    0x4009,  //cell height
    0x0014,  //turn off dword mode
    0xe317   //turn on byte mode
  }
};

static TXSetup X320Y480 = {
  320,
  480,
  0x0e3,
  {
    0x0d06,  //vertical total
    0x3e07,  //overflow (bit 8 of vertical counts)
    0x4009,  //cell height (2 to double-scan)
    0xea10,  //v sync start
    0xac11,  //v sync end and protect cr0-cr7
    0xdf12,  //vertical displayed
    0x0014,  //turn off dword mode
    0xe715,  //v blank start
    0x0616,  //v blank end
    0xe317   //turn on byte mode
  }
};

static TXSetup X360Y400 = {
  360,
  400,
  0x0e7,
  {
    0x6b00,  //horz total
    0x5901,  //horz displayed
    0x5a02,  //start horz blanking
    0x8e03,  //end horz blanking
    0x5e04,  //start h sync
    0x8a05,  //end h sync
    0x4009,  //cell height
    0x0014,  //turn off dword mode
    0xe317   //turn on byte mode
  }
};

static TXSetup X360Y480 = {
  360,
  480,
  0x0e7,
  {
    0x6b00,  //horz total
    0x5901,  //horz displayed
    0x5a02,  //start horz blanking
    0x8e03,  //end horz blanking
    0x5e04,  //start h sync
    0x8a05,  //end h sync
    0x0d06,  //vertical total
    0x3e07,  //overflow
    0x4009,  //cell height
    0xea10,  //v sync start
    0xac11,  //v sync end and protect cr0-cr7
    0xdf12,  //vertical displayed
    0x2d13,  //offset
    0x0014,  //turn off dword mode
    0xe715,  //v blank start
    0x0616,  //v blank end
    0xe317   //turn on byte mode
  }
};

static TXSetup X360Y360 = {
  360,
  360,
  0x0e7,
  {
    0x6b00,  //horz total
    0x5901,  //horz displayed
    0x5a02,  //start horz blanking
    0x8e03,  //end horz blanking
    0x5e04,  //start h sync
    0x8a05,  //end h sync
    0x4009,  //cell height
    0x8810,  //v sync start
    0x8511,  //v sync end and protect cr0-cr7
    0x6712,  //vertical displayed
    0x2d13,  //offset
    0x0014,  //turn off dword mode
    0x6d15,  //v blank start
    0xba16,  //v blank end
    0xe317   //turn on byte mode
  }
};

static TXSetup X376Y308 = {
  376,
  308,
  0x0e7,
  {
    0x6e00,  //horz total
    0x5d01,  //horz displayed
    0x5e02,  //start horz blanking
    0x9103,  //end horz blanking
    0x6204,  //start h sync
    0x8f05,  //end h sync
    0x6206,  //vertical total
    0x0f07,  //overflow
    0x4009,  //Unknown?-->VPutz
    0x310f,  //Unknown?-->VPutz
    0x3710,  //v sync start
    0x8911,  //v sync end and protect cr0-cr7
    0x3312,  //vertical displayed
    0x2f13,  //offset
    0x0014,  //turn off dword mode
    0x3c15,  //v blank start
    0x5c16,  //v blank end
    0xe317   //turn on byte mode
  }
};

static TXSetup X376Y564 = {
  376,
  564,
  0x0e7,
  {
    0x6e00,  //horz total
    0x5d01,  //horz displayed
    0x5e02,  //start horz blanking
    0x9103,  //end horz blanking
    0x6204,  //start h sync
    0x8f05,  //end h sync
    0x6206,  //vertical total
    0xf007,  //overflow
    0x6009,  //Unknown?-->VPutz
    0x310f,  //Unknown?-->VPutz
    0x3710,  //v sync start
    0x8911,  //v sync end and protect cr0-cr7
    0x3312,  //vertical displayed
    0x2f13,  //offset
    0x0014,  //turn off dword mode
    0x3c15,  //v blank start
    0x5c16,  //v blank end
    0xe317   //turn on byte mode
  }
};

/*
static TXSetup X256Y400 = {
  256,
  400,
  0x0e3,
  {
    0x5f00,
    0x3f01,
    0x4202,
    0x9f03,
    0x4c04,
    0x0005,
    0x4009,
    0x0014,
    0xE317,
  }
};


static TXSetup X256Y480 = {
  256,
  480,
  0x0e3,
  {
    0x5f00,
    0x3f01,
    0x4202,
    0x9f03,
    0x4c04,
    0x0005,
    0x0d06,
    0x3e07,
    0x4009,
    0xea10,
    0xac11,
    0xdf12,
    0x0014,
    0xe715,
    0x0616,
    0xe317
  }
};
*/

static TXSetup X360Y270 = {
  360,
  270,
  0x0e7,
  {
    0x12e7,
    0x6b00,
    0x5901,
    0x5a02,
    0x8e03,
    0x5e04,
    0x8a05,
    0x3006,
    0xf007,
    0x0008,
    0x6109,
    0x2010,
    0xa911,
    0x1b12,
    0x2d13,
    0x0014,
    0x1f15,
    0x2f16,
    0xe317
  }
};

static TXSetup X400Y300 = {
  400,
  300,
  0x0e7,
  {
    0x7100,
    0x6301,
    0x6402,
    0x9203,
    0x6504,
    0x8205,
    0x4606,
    0x1f07,
    0x0008,
    0x4009,
    0x3110,
    0x8011,
    0x2b12,
    0x3213,
    0x0014,
    0x2f15,
    0x4416,
    0xe317,
  }
};


static TXSetup * xSetup [] = {
  // 2+ page - any vga videocards - double buffer
  //           (virtual screen set to maxX+16*maxX+16)
  // 1 page  - 100% vga compatible videocards - single buffer
  &X256Y200,
  &X256Y224,
  &X256Y240,
  &X256Y256,
  &X256Y400,
  &X256Y480, // 1 page
  &X320Y200,
  &X320Y240,
  &X320Y400, // 1 page
  &X320Y480, // 1 page
  &X360Y200,
  &X360Y240,
  &X360Y270,
  &X360Y360, // 1 page
  &X360Y400, // 1 page
  &X360Y480, // 1 page
  &X376Y282,
  &X376Y308,
  // not scrolling xmode for all vga's
  &X376Y564,

  &X400Y300,
};

#define LAST_X_MODE (0x13+1)
